
''' Tinymovr endpoints definitions module.

This module includes definitions for all Tinymovr endpoints .

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
'''

from tinymovr.iface import DataType

Endpoints = {
    "nmt":
    {
        "description": "CANOpen NMT Message",
        "type": "w",
        "ep_id": 0x000
    },
    "heartbeat":
    {
        "description": "CANOpen Heartbeat Message",
        "type": "w",
        "ep_id": 0x700
    },
    "estop":
    {
        "description": "Tinymovr Estop Message",
        "type": "w",
        "ep_id": 0x002
    },
    "state":
    {
        "description": "Get Tinymovr State, Control Mode and Error Code",
        "type": "r",
        "ep_id": 0x003,
        "types": (DataType.UINT8, DataType.UINT8, DataType.UINT8),
        "labels": ("error", "state", "mode")
    },
    "can_config":
    {
        "description": "Get CAN Config",
        "type": "r",
        "ep_id": 0x005,
        "types": (DataType.UINT8, DataType.UINT16),
        "labels": ("id", "baud_rate")
    },
    "set_can_config":
    {
        "description": "Set CAN Config",
        "type": "w",
        "ep_id": 0x006,
        "types": (DataType.UINT8, DataType.UINT16),
        "defaults": (0,),
        "labels": ("id", "baud_rate")
    },
    "set_state":
    {
        "description": "Set Tinymovr State and optionally Control Mode",
        "type": "w",
        "ep_id": 0x007,
        "types": (DataType.UINT8, DataType.UINT8),
        "defaults": (0,),
        "labels": ("state", "mode")
    },
    "encoder_estimates":
    {
        "description": "Get Encoder Estimates (Position, Velocity)",
        "type": "r",
        "ep_id": 0x009,
        "types": (DataType.FLOAT, DataType.FLOAT),
        "labels": ("position", "velocity")
    },
    "set_pos_setpoint":
    {
        "description": "Set Position Setpoint",
        "type": "w",
        "ep_id": 0x00C,
        "types": (DataType.FLOAT, DataType.INT16, DataType.INT16),
        "defaults": (0, 0),
        "labels": ("position", "velocity_ff", "current_ff")
    },
    "set_vel_setpoint":
    {
        "description": "Set Velocity Setpoint",
        "type": "w",
        "ep_id": 0x00D,
        "types": (DataType.FLOAT, DataType.FLOAT),
        "defaults": (0,),
        "labels": ("velocity", "current_ff")
    },
    "set_cur_setpoint":
    {
        "description": "Set Current Setpoint",
        "type": "w",
        "ep_id": 0x00E,
        "types": (DataType.FLOAT,)
    },
    "set_limits":
    {
        "description": "Set Limits (Velocity, Current)",
        "type": "w",
        "ep_id": 0x00F,
        "types": (DataType.FLOAT, DataType.FLOAT),
        "labels": ("velocity", "current")
    },
    "Iq":
    {
        "description": "Get current setpoint and estimate",
        "type": "r",
        "ep_id": 0x014,
        "types": (DataType.FLOAT, DataType.FLOAT),
        "labels": ("setpoint", "estimate")
    },
    "limits":
    {
        "description": "Get velocity and current limits",
        "type": "r",
        "ep_id": 0x015,
        "types": (DataType.FLOAT, DataType.FLOAT),
        "labels": ("velocity", "current")
    },
    "reset":
    {
        "description": "Tinymovr Reset Message",
        "type": "w",
        "ep_id": 0x016
    },
    "Vbus":
    {
        "description": "Get Bus Voltage",
        "type": "r",
        "ep_id": 0x017,
        "types": (DataType.FLOAT,)
    },
    "gains":
    {
        "description": "Get Gains (Position, Velocity)",
        "type": "r",
        "ep_id": 0x018,
        "types": (DataType.FLOAT, DataType.FLOAT),
        "labels": ("position", "velocity")
    },
    "set_gains":
    {
        "description": "Set Gains (Position, Velocity)",
        "type": "w",
        "ep_id": 0x019,
        "types": (DataType.FLOAT, DataType.FLOAT),
        "labels": ("position", "velocity")
    },
    "device_info":
    {
        "description": "Get Device Info",
        "type": "r",
        "ep_id": 0x01A,
        "types": (DataType.UINT32, DataType.UINT16, DataType.UINT16),
        "labels": ("device_id", "temp", "encoder_cpr")
    },
    "timings":
    {
        "description": "Get Processor Timings (Busy Cycles/PWM, Total Cycles/PWM)",
        "type": "r",
        "ep_id": 0x01B,
        "types": (DataType.UINT32, DataType.UINT32),
        "labels": ("busy", "total")
    },
    "save_config":
    {
        "description": "Save Configuration",
        "type": "w",
        "ep_id": 0x01C
    },
    "erase_config":
    {
        "description": "Erase Configuration",
        "type": "w",
        "ep_id": 0x01D
    },
    "motor_info":
    {
        "description": "Get Attached Motor Info (Calibrated, Resistance, Pole Pairs, Inductance)",
        "type": "r",
        "ep_id": 0x01E,
        "types": (DataType.UINT8, DataType.UINT16, DataType.UINT8, DataType.FLOAT),
        "labels": ("calibrated", "R", "pole_pairs", "L")
    },
}



from enum import Enum

class ErrorIDs(Enum):
    NoError = 0
    InvalidState = 1
    IllegalValue = 2
    VBusUndervoltage = 3
    OverCurrent = 4
    PhaseResistanceOutOfRange = 5
    PhaseInductanceOutOfRange = 6
    InvalidPolePairs = 7

error_descriptions = {
    ErrorIDs.NoError: "No error",
    ErrorIDs.InvalidState: "Attempt to transition to invalid state",
    ErrorIDs.IllegalValue: "Illegal value passed to endpoint",
    ErrorIDs.VBusUndervoltage: "Power supply bus under low voltage threshold",
    ErrorIDs.OverCurrent: "Phase current exceeded threshold",
    ErrorIDs.PhaseResistanceOutOfRange: "Measured phase resistance is out of range",
    ErrorIDs.PhaseResistanceOutOfRange: "Measured phase inductance is out of range",
    ErrorIDs.InvalidPolePairs: "Pole pairs calculation did not converge to a valid value"
}    


from tinymovr.errors import ErrorIDs, error_descriptions
from tinymovr.endpoints import Endpoints
from tinymovr.tinymovr import Tinymovr
from tinymovr.user_wrapper import UserWrapper
from tinymovr.shell import spawn_shell



''' Tinymovr user wrapper module.

This module includes a friendly Tinymovr user wrapper class that makes it
easier for the user to issue commands to Tinymovr through the shell.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
'''

from tinymovr import Tinymovr, ErrorIDs, error_descriptions

class UserWrapper:
    '''
    This is a user-friendly wrapper to the Tinymover class, that
    allows more intuitive functions to be used instead of plain
    getters/setters.
    This class sacrifices Pythonic-ness (e.g. raising exceptions)
    for better user-friendliness (e.g. using simple print statements).
    There are also additional checks performed prior to the 
    '''
    def __init__(self, **kwargs):
        self.tinymovr = Tinymovr(**kwargs)
        self._encoder_cpr = -1

    def __getattr__(self, attr):
        return getattr(self.tinymovr, attr)

    def calibrate(self):
        '''
        Enter calibration state, perform motor and encoder calibration
        '''
        state = self.tinymovr.state
        if state.error != 0:
            print("Error flag present, cannot continue with calibration. Please reset Tinymovr.")
        elif state.state != 0:
            print("Tinymovr state is not idle, calibration needs to be started from idle state.")
        else:
            input("Ready to calibrate. Please remove any loads from the motor and hit Enter to continue")
            self.tinymovr.calibrate()

    def idle(self):
        '''
        Enter idle state
        '''
        self.tinymovr.set_state(0)

    def position_control(self):
        '''
        Enter closed loop control state and position control mode
        '''
        state = self.tinymovr.state
        if state.error != 0:
            print("Error flag present, cannot enable position control. Please reset Tinymovr.")
        elif state.mode == 1:
            print("Tinymovr is currently calibrating, please do not interrupt.")
        else:
            self.tinymovr.position_control()

    def velocity_control(self):
        '''
        Enter closed loop control state and velocity control mode
        '''
        state = self.tinymovr.state
        if state.error != 0:
            print("Error flag present, cannot enable velocity control. Please reset Tinymovr.")
        elif state.mode == 1:
            print("Tinymovr is currently calibrating, please do not interrupt.")
        else:
            self.tinymovr.velocity_control()

    def current_control(self):
        '''
        Enter closed loop control state and current control mode
        '''
        state = self.tinymovr.state
        if state.error != 0:
            print("Error flag present, cannot enable current control. Please reset Tinymovr.")
        elif state.mode == 1:
            print("Tinymovr is currently calibrating, please do not interrupt.")
        else:
            self.tinymovr.current_control()

    def get_velocity_estimate_rpm(self, value):
        '''
        Get velocity estimate in RPM by automatically
        converting from encoder counts per second.

        NOTE: This convenience function exists only for
        velocity control mode and does not support
        current feed-forward value.
        '''
        return 60 * (self.tinymovr.encoder_estimates.velocity / self.encoder_cpr)

    def set_velocity_setpoint_rpm(self, value):
        '''
        Set velocity setpoint in RPM by automatically
        converting to encoder counts per second

        NOTE: This convenience function exists only for
        velocity control mode and does not support
        current feed-forward value.
        '''
        self.tinymovr.set_velocity_setpoint(value * self.encoder_cpr / 60)

    def error(self):
        '''
        Report controller error in human-readable form
        '''
        state = self.tinymovr.state
        error_id = ErrorIDs(state.error)
        print(error_descriptions[error_id] + " (error code: " + str(error_id) + ")")

    @property
    def encoder_cpr(self):
        if self._encoder_cpr < 2048:
            self._encoder_cpr = self.tinymovr.device_info.encoder_cpr
        assert(self._encoder_cpr >= 2048)
        return self._encoder_cpr

    def __dir__(self):
        tm_keys = self.tinymovr.__dir__()
        self_keys = object.__dir__(self)
        self_keys.remove("tinymovr")
        return tm_keys + self_keys



#!/usr/bin/env python3
"""Tinymovr Shell Utility

Usage:
    tinymovr [--iface=<iface>]
    tinymovr -h | --help
    tinymovr --version

Options:
    --iface=<iface>  CAN interface to use [default: arduino_can].
"""

'''
This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
'''

import logging
import pkg_resources
import can
import IPython

from docopt import docopt

from tinymovr import UserWrapper
from tinymovr.iface import CAN

shell_name = 'Tinymovr Shell Utility'
base_name = "tm"

def spawn_shell():
    version = pkg_resources.require("tinymovr")[0].version
    arguments = docopt(__doc__, version=shell_name + ' ' + str(version))
    logging.getLogger('parso').setLevel(logging.WARNING)
    logging.getLogger('asyncio').setLevel(logging.WARNING)
    can.util.set_logging_level('warning')

    logger = logging.getLogger('tinymovr')
    logger.setLevel(logging.DEBUG)
    iface = CAN(can.interface.Bus(interface=arguments['--iface'], channel='can0'))
    tms = {}
    for node_id in range(1, 9):
        try:
            tm = UserWrapper(node_id=node_id, iface=iface)
            _ = tm.device_info
            tm_string = base_name+str(node_id)
            logger.info("Connected to " + tm_string)
            tms[tm_string] = tm
        except TimeoutError:
            logger.error("Node " + str(node_id) + " timed out")
        except IOError:
            logger.error("Node " + str(node_id) + " received abnormal message (possibly wrong ID?)")
    if len(tms) == 0:
        logger.error("No Tinymovr instances detected. Exiting shell...")
    else:
        tms["tms"] = list(tms.values())
        print(shell_name + ' ' + str(version))
        print("Access Tinymovr instances as tmx, where x is the index, e.g. tm1")
        print("Instances are also available by index in the tms list.")
        IPython.start_ipython(argv=["--no-banner"], user_ns=tms)
        logger.debug("Exiting shell...")

if __name__ == "__main__":
    main()




''' Tinymovr base module.

This module includes the base Tinymovr class that implements the API
to interface with the Tinymovr motor control board.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.
This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>.
'''

import copy
from tinymovr.iface import CAN, CANBusCodec, DataType
from tinymovr import Endpoints
from tinymovr.attr_object import AttrObject

class Tinymovr:

    def __init__(self, node_id, iface, codec=CANBusCodec(), eps=Endpoints):
        self.node_id = node_id
        self.iface = iface
        self.codec = codec
        self.endpoints = eps

    def __getattr__(self, attr):
        if attr in self.endpoints:
            d = self.endpoints[attr]
            if d["type"] == "w":
                def wrapper(*args, **kwargs):
                    f_args = copy.copy(args)
                    if "types" in d:
                        slack = len(d["types"]) - len(f_args)
                        if slack > 0:
                            slack_defaults = d["defaults"][-slack:]
                            f_args = f_args + slack_defaults
                        payload = self.codec.serialize(f_args, *d["types"])
                        self.iface.send_new(self.node_id, d["ep_id"], payload=payload)
                    else:
                        self.iface.send_new(self.node_id, d["ep_id"])
                return wrapper
            elif d["type"] == "r":
                self.iface.send_new(self.node_id, d["ep_id"], rtr=True)
                payload = self.iface.receive(self.node_id, d["ep_id"], 1)
                values = self.codec.deserialize(payload, *d["types"])
                if len(values) == 1:
                    return values[0]
                else:
                    assert("labels" in d)
                    return AttrObject(d["labels"], values)

    def __dir__(self):
        return list(self.endpoints.keys())

    def calibrate(self):
        self.set_state(1)

    def idle(self):
        self.set_state(0)

    def position_control(self):
        self.set_state(2, 2)

    def velocity_control(self):
        self.set_state(2, 1)

    def current_control(self):
        self.set_state(2, 0)



import pathlib
import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setuptools.setup(
    name="tinymovr",
    version="0.1.0",
    author="Yannis Chatzikonstantinou",
    author_email="info@tinymovr.com",
    description="Tinymovr Studio",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/yconst/Tinymovr",
    packages=['tinymovr'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "ipython",
        "python-can",
        "pyserial",
        "docopt"
    ],
    entry_points={
        'can.interface': [
            "arduino_can=tinymovr.bus:ArduinoCAN",
        ],
        "console_scripts": [
            "tinymovr=tinymovr.shell:spawn_shell"
        ],
    }
    
)

